/*
FILE:     TTable.C
*/
/*
VERSION:  8.003
*/
/*
DATE:     4/24/2007
*/
/*
  Comments and Questions to: sw-help@rcsb.rutgers.edu
*/
/*
COPYRIGHT 1999-2007 Rutgers - The State University of New Jersey

This software is provided WITHOUT WARRANTY OF MERCHANTABILITY OR
FITNESS FOR A PARTICULAR PURPOSE OR ANY OTHER WARRANTY, EXPRESS OR
IMPLIED.  RUTGERS MAKE NO REPRESENTATION OR WARRANTY THAT THE
SOFTWARE WILL NOT INFRINGE ANY PATENT, COPYRIGHT OR OTHER
PROPRIETARY RIGHT.

The user of this software shall indemnify, hold harmless and defend
Rutgers, its governors, trustees, officers, employees, students,
agents and the authors against any and all claims, suits,
losses, liabilities, damages, costs, fees, and expenses including
reasonable attorneys' fees resulting from or arising out of the
use of this software.  This indemnification shall include, but is
not limited to, any and all claims alleging products liability.
*/
/*
               PDB SOFTWARE LICENSE AGREEMENT

BY CLICKING THE ACCEPTANCE BUTTON OR INSTALLING OR USING 
THIS "SOFTWARE, THE INDIVIDUAL OR ENTITY LICENSING THE  
SOFTWARE ("LICENSEE") IS CONSENTING TO BE BOUND BY AND IS 
BECOMING A PARTY TO THIS AGREEMENT.  IF LICENSEE DOES NOT 
AGREE TO ALL OF THE TERMS OF THIS AGREEMENT
THE LICENSEE MUST NOT INSTALL OR USE THE SOFTWARE.

1. LICENSE AGREEMENT

This is a license between you ("Licensee") and the Protein Data Bank (PDB) 
at Rutgers, The State University of New Jersey (hereafter referred to 
as "RUTGERS").   The software is owned by RUTGERS and protected by 
copyright laws, and some elements are protected by laws governing 
trademarks, trade dress and trade secrets, and may be protected by 
patent laws. 

2. LICENSE GRANT

RUTGERS grants you, and you hereby accept, non-exclusive, royalty-free 
perpetual license to install, use, modify, prepare derivative works, 
incorporate into other computer software, and distribute in binary 
and source code format, or any derivative work thereof, together with 
any associated media, printed materials, and on-line or electronic 
documentation (if any) provided by RUTGERS (collectively, the "SOFTWARE"), 
subject to the following terms and conditions: (i) any distribution 
of the SOFTWARE shall bind the receiver to the terms and conditions 
of this Agreement; (ii) any distribution of the SOFTWARE in modified 
form shall clearly state that the SOFTWARE has been modified from 
the version originally obtained from RUTGERS.  

2. COPYRIGHT; RETENTION OF RIGHTS.  

The above license grant is conditioned on the following: (i) you must 
reproduce all copyright notices and other proprietary notices on any 
copies of the SOFTWARE and you must not remove such notices; (ii) in 
the event you compile the SOFTWARE, you will include the copyright 
notice with the binary in such a manner as to allow it to be easily 
viewable; (iii) if you incorporate the SOFTWARE into other code, you 
must provide notice that the code contains the SOFTWARE and include 
a copy of the copyright notices and other proprietary notices.  All 
copies of the SOFTWARE shall be subject to the terms of this Agreement.  

3. NO MAINTENANCE OR SUPPORT; TREATMENT OF ENHANCEMENTS 

RUTGERS is under no obligation whatsoever to: (i) provide maintenance 
or support for the SOFTWARE; or (ii) to notify you of bug fixes, patches, 
or upgrades to the features, functionality or performance of the 
SOFTWARE ("Enhancements") (if any), whether developed by RUTGERS 
or third parties.  If, in its sole discretion, RUTGERS makes an 
Enhancement available to you and RUTGERS does not separately enter 
into a written license agreement with you relating to such bug fix, 
patch or upgrade, then it shall be deemed incorporated into the SOFTWARE 
and subject to this Agreement. You are under no obligation whatsoever 
to provide any Enhancements to RUTGERS or the public that you may 
develop over time; however, if you choose to provide your Enhancements 
to RUTGERS, or if you choose to otherwise publish or distribute your 
Enhancements, in source code form without contemporaneously requiring 
end users or RUTGERS to enter into a separate written license agreement 
for such Enhancements, then you hereby grant RUTGERS a non-exclusive,
royalty-free perpetual license to install, use, modify, prepare
derivative works, incorporate into the SOFTWARE or other computer
software, distribute, and sublicense your Enhancements or derivative
works thereof, in binary and source code form.

4. FEES.  There is no license fee for the SOFTWARE.  If Licensee
wishes to receive the SOFTWARE on media, there may be a small charge
for the media and for shipping and handling.  Licensee is
responsible for any and all taxes.

5. TERMINATION.  Without prejudice to any other rights, Licensor
may terminate this Agreement if Licensee breaches any of its terms
and conditions.  Upon termination, Licensee shall destroy all
copies of the SOFTWARE.

6. PROPRIETARY RIGHTS.  Title, ownership rights, and intellectual
property rights in the Product shall remain with RUTGERS.  Licensee 
acknowledges such ownership and intellectual property rights and will 
not take any action to jeopardize, limit or interfere in any manner 
with RUTGERS' ownership of or rights with respect to the SOFTWARE.  
The SOFTWARE is protected by copyright and other intellectual 
property laws and by international treaties.  Title and related 
rights in the content accessed through the SOFTWARE is the property 
of the applicable content owner and is protected by applicable law.  
The license granted under this Agreement gives Licensee no rights to such
content.

7. DISCLAIMER OF WARRANTY.  THE SOFTWARE IS PROVIDED FREE OF 
CHARGE, AND, THEREFORE, ON AN "AS IS" BASIS, WITHOUT WARRANTY OF 
ANY KIND, INCLUDING WITHOUT LIMITATION THE WARRANTIES THAT IT 
IS FREE OF DEFECTS, MERCHANTABLE, FIT FOR A PARTICULAR PURPOSE 
OR NON-INFRINGING.  THE ENTIRE RISK AS TO THE QUALITY AND 
PERFORMANCE OF THE SOFTWARE IS BORNE BY LICENSEE.  SHOULD THE 
SOFTWARE PROVE DEFECTIVE IN ANY RESPECT, THE LICENSEE AND NOT 
LICENSOR ASSUMES THE ENTIRE COST OF ANY SERVICE AND REPAIR.  
THIS DISCLAIMER OF WARRANTY CONSTITUTES AN ESSENTIAL PART OF 
THIS AGREEMENT.  NO USE OF THE PRODUCT IS AUTHORIZED HEREUNDER 
EXCEPT UNDER THIS DISCLAIMER.

8. LIMITATION OF LIABILITY.  TO THE MAXIMUM EXTENT PERMITTED BY
APPLICABLE LAW,  IN NO EVENT WILL LICENSOR BE LIABLE FOR ANY 
INDIRECT, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES ARISING 
OUT OF THE USE OF OR INABILITY TO USE THE SOFTWARE, INCLUDING, 
WITHOUT LIMITATION, DAMAGES FOR LOSS OF GOODWILL, WORK 
STOPPAGE, COMPUTER FAILURE OR MALFUNCTION, OR ANY AND ALL 
OTHER COMMERCIAL DAMAGES OR LOSSES, EVEN IF ADVISED OF THE
POSSIBILITY THEREOF. 
*/


/*!
** \file TTable.C
**
** \brief Implementation file for TTable class. 
*/


#include <stdexcept>
#include <string>
#include <iostream>
#include <iomanip>

#include "Exceptions.h"
#include "GenString.h"
#include "TTable.h"


using namespace std;


TTable::TTable() : _numRows(0)
#ifndef IS_TABLE_ROW_DELETE_AS_REMOVE
  , _numDelRows(0)
#endif
{


}


TTable::TTable(const TTable& inTable)
{

    _numRows = inTable._numRows;

    // This statement only copies pointer addresses and not the column
    // vectors referenced by those pointers.
    _data = inTable._data;

    // Here, column vectors themselves are copied and previously set
    // pointers are properly initialized.

    for (unsigned int colIndex = 0; colIndex < _data.size(); ++colIndex)
    {
        _data[colIndex] = new vector<string>(*(inTable._data[colIndex]));
    }

#ifndef IS_TABLE_ROW_DELETE_AS_REMOVE
    _rowMap = inTable._rowMap;
    _numDelRows = inTable._numDelRows;
#endif

}


TTable::~TTable()
{

    Clear();

}


TTable& TTable::operator=(const TTable& inTable)
{

    if (this != &inTable)
    {
        Clear();

        _numRows = inTable._numRows;

        // This statement only copies pointer addresses and not the column
        // vectors referenced by those pointers.
        _data = inTable._data;

        // Here, column vectors themselves are copied and previously set
        // pointers are properly initialized.

        for (unsigned int colIndex = 0; colIndex < _data.size();
          ++colIndex)
        {
            _data[colIndex] = new vector<string>(*(inTable._data[colIndex]));
        }
    }

#ifndef IS_TABLE_ROW_DELETE_AS_REMOVE
    _rowMap = inTable._rowMap;
    _numDelRows = inTable._numDelRows;
#endif

    return(*this);

}


void TTable::Clear()
{

    if (!_data.empty())
    {
        for (unsigned int colIndex = 0; colIndex < _data.size(); colIndex++)
        {
            _data[colIndex]->clear();
            delete _data[colIndex];
        }

        _data.clear();
    }

    _numRows = 0;

#ifndef IS_TABLE_ROW_DELETE_AS_REMOVE
    _rowMap.clear();
    _numDelRows = 0;
#endif

}


unsigned int TTable::AddColumn(const vector<string>& col)
{

    InsertColumn(_data.size(), col);

    return(_data.size());

}


void TTable::InsertColumn(const unsigned int colIndex,
  const vector<string>& col)
{

    if (colIndex > _data.size())
    {
        // Wrong column index.
        throw out_of_range("Invalid column index in TTable::InsertColumn");
    }

    if ((GetNumRows() != 0) && (col.size() > GetNumRows()))
    {
        // Wrong column index.
        throw out_of_range("Invalid column size in TTable::InsertColumn");
    }

    CreateColumn(colIndex, col);

}


void TTable::FillColumn(const unsigned int colIndex, const vector<string>& col,
  const unsigned int fromRowIndex)
{

    if (colIndex >= _data.size())
    {
        // Wrong column index.
        throw out_of_range("Invalid column index in TTable::FillColumn");
    }

    if (col.empty())
    {
        return;
    }

    if (fromRowIndex > GetNumRows())
    {
        // Wrong from row index.
        throw out_of_range("Invalid from row index in TTable::FillColumn");
    }

    if ((GetNumRows() != 0) && (col.size() > GetNumRows() - fromRowIndex))
    {
        // Wrong column index.
        throw out_of_range("Invalid column size in TTable::FillColumn");
    }

    copy(col.begin(), col.end(), _data[colIndex]->begin() +
      IntRowIndex(fromRowIndex));

}


void TTable::GetColumn(vector<string>& col, const unsigned int colIndex,
      const unsigned int fromRowIndex, unsigned int toRowIndex)
{

    // VLAD - ERROR HANDLING - ADD

    col.clear();

    for (unsigned int rowI = fromRowIndex; rowI < toRowIndex; ++rowI)
    {
        col.push_back(operator()(rowI, colIndex));
    }

}


void TTable::ClearColumn(const unsigned int colIndex)
{

    if (colIndex >= _data.size())
    {
        // Wrong column index.
        throw out_of_range("Invalid column index in TTable::ClearColumn");
    }

    for (unsigned int rowIndex = 0; rowIndex < _data[colIndex]->size();
      ++rowIndex)
    {
        (*_data[colIndex])[rowIndex].clear();
    }

}


void TTable::DeleteColumn(const unsigned int colIndex)
{

    if (colIndex >= _data.size())
    {
        // Wrong column index.
        throw out_of_range("Invalid column index in TTable::DeleteColumn");
    }

#ifndef IS_TABLE_ROW_DELETE_AS_REMOVE
    unsigned int oldNumRows = _numRows;
#endif

    _data[colIndex]->clear();

    delete _data[colIndex];

    _data.erase(_data.begin() + colIndex);

    // Determines the maximum size of all columns and sets _numRows
    for (unsigned int colIndex = 0; colIndex < _data.size(); ++colIndex)
    {
        if (_data[colIndex]->size() > _numRows)
        {
            _numRows = _data[colIndex]->size();
        }
    }

#ifndef IS_TABLE_ROW_DELETE_AS_REMOVE
    if (oldNumRows > _numRows)
    {
        // If deleting of column deleted some rows at the end, remove them
        // from row map.
        ReduceRowMap(oldNumRows - _numRows);
    }
#endif

}


unsigned int TTable::AddRow(const vector<string>& row)
{

    return(InsertRow(GetNumRows(), row));

}


unsigned int TTable::InsertRow(const unsigned int atRowIndex,
  const vector<string>& row)
{

    if (atRowIndex > GetNumRows())
    {
        // Wrong row index.
        throw out_of_range("Invalid row index in STable::InsertRow");
    }

    if ((!_data.empty()) && (row.size() > _data.size()))
    {
        throw out_of_range("Invalid row size in STable::InsertRow");
    }

    if ((GetNumRows() == 0) && (row.size() > _data.size()))
    {
        unsigned int numCols = row.size() - _data.size();
        for (unsigned int colI = 0; colI < numCols; ++colI)
        {
            _data.push_back(new vector<string>);
        }
    }

#ifndef IS_TABLE_ROW_DELETE_AS_REMOVE
    if (_numDelRows != 0)
    {
        UnMarkRowDeleted(atRowIndex);
    }
    else
    {
        EnlargeRowMap(1);
#endif
        _numRows++;

        unsigned int intRowIndex = IntRowIndex(atRowIndex);

        for (unsigned int colIndex = 0; colIndex < _data.size(); ++colIndex)
        {
            _data[colIndex]->insert(_data[colIndex]->begin() + intRowIndex,
              1, String::Empty);
        }

#ifndef IS_TABLE_ROW_DELETE_AS_REMOVE
    }
#endif

    FillRow(atRowIndex, row);

    return(GetNumRows());

}


void TTable::FillRow(const unsigned int rowIndex, const vector<string>& row,
  const unsigned int fromColIndex)
{

    if (rowIndex >= GetNumRows())
    {
        // Wrong row index.
        throw out_of_range("Invalid row index in STable::FillRow");
    }

    if (row.empty())
    {
        return;
    }

    if (row.size() > _data.size() - fromColIndex)
    {
        throw out_of_range("Invalid row size in TTable::FillRow");
    }

    unsigned int intRowIndex = IntRowIndex(rowIndex);

    for (unsigned int colIndex = 0; colIndex < row.size(); ++colIndex)
    {
        (*_data[fromColIndex + colIndex])[intRowIndex] =
          row[colIndex];
    }

}


void TTable::ClearRow(const unsigned int rowIndex)
{

    if (rowIndex >= GetNumRows())
    {
        // Wrong row index.
        throw out_of_range("Invalid row index in STable::ClearRow");
    }

    unsigned int intRowIndex = IntRowIndex(rowIndex);

    for (unsigned int colIndex = 0; colIndex < _data.size(); ++colIndex)
    {
        (*_data[colIndex])[intRowIndex].clear();
    }

}


void TTable::DeleteRow(const unsigned int rowIndex)
{

    if (rowIndex >= GetNumRows())
    {
        // Wrong row index.
        throw out_of_range("Invalid row index in STable::DeleteRow");
    }

#ifdef IS_TABLE_ROW_DELETE_AS_REMOVE
    for (unsigned int colIndex = 0; colIndex < _data.size(); ++colIndex)
    {
        _data[colIndex]->erase(_data[colIndex]->begin() + rowIndex);
    }
#endif

#ifndef IS_TABLE_ROW_DELETE_AS_REMOVE
    MarkRowDeleted(rowIndex);
#else
    _numRows--;
#endif

}


ostream& operator<<(ostream& out, const TTable& tTable)
{

    for (unsigned int rowI = 0; rowI < tTable.GetNumRows(); ++rowI)
    {
        for (unsigned int colI = 0; colI < tTable.GetNumColumns(); ++colI)
        {
            out << setw(10) << tTable(rowI, colI) << " ";
        }

        out << endl;
    }

    return(out);

}


void TTable::CreateColumn(const unsigned int colIndex,
  const vector<string>& col)
{

    if (colIndex > _data.size())
    {
        // Wrong column index.
        throw out_of_range("Invalid column index in TTable::CreateColumn");
    }

    if ((GetNumRows() != 0) && (col.size() > GetNumRows()))
    {
        // Wrong column size.
        throw out_of_range("Invalid column size in TTable::CreateColumn");
    }

    if ((!col.empty()) && (colIndex != 0) && (_data[colIndex - 1]->empty()))
    {
        // Previous column empty. Cannot insert, since rectangularity will be
        // violated.
        throw out_of_range("Previous column empty TTable::CreateColumn");
    }

    if (_numRows == 0)
    {
        // For empty table, new column defines the initial number of rows.
        _numRows = col.size();
#ifndef IS_TABLE_ROW_DELETE_AS_REMOVE
        EnlargeRowMap(_numRows);
#endif
    }

    _data.insert(_data.begin() + colIndex, new vector<string>);

    if (_data[colIndex]->empty())
    {
        // If table column empty, insert empty cells.
        _data[colIndex]->insert(_data[colIndex]->begin(), _numRows,
          String::Empty);
    }

    copy(col.begin(), col.end(), _data[colIndex]->begin());

}


void TTable::GetRow(vector<string>& row, const unsigned int rowIndex,
  const unsigned int fromColIndex, unsigned int toColIndex)
{

    row.clear();

    if (rowIndex >= GetNumRows())
    {
        // Wrong row index.
        throw out_of_range("Invalid row index in TTable::GetRow");
    }

    if (fromColIndex >= toColIndex)
    {
        // To index less than from index.
        throw out_of_range("Invalid column index range in TTable::GetRow");
    }

    if (fromColIndex >= _data.size())
    {
        // Wrong from column index.
        throw out_of_range("Invalid from column index in TTable::GetRow");
    }

    if (toColIndex > _data.size())
    {
        // Wrong to column index.
        throw out_of_range("Invalid to column index in TTable::GetRow");
    }

    // toColIndex is inclusive, so one more spot for it needs to be reserved
    row.reserve(toColIndex - fromColIndex);

    unsigned int intRowIndex = IntRowIndex(rowIndex);

    for (unsigned int colIndex = fromColIndex; colIndex < toColIndex;
      colIndex++)
    {
        row.push_back((*_data[colIndex])[intRowIndex]);
    }

}


string& TTable::operator()(const unsigned int rowIndex,
      const unsigned int colIndex)
{

    if (colIndex >= _data.size())
    {
        // Wrong column index.
        throw out_of_range("Invalid column index in TTable::UpdateCell");
    }

    if (rowIndex >= GetNumRows())
    {
        // Wrong row index.
        throw out_of_range("Invalid row index in TTable::UpdateCell");
    }

    return((*_data[colIndex])[IntRowIndex(rowIndex)]);

}


const string& TTable::operator()(const unsigned int rowIndex, 
      const unsigned int colIndex) const
{

    if (colIndex >= _data.size())
    {
        // Wrong column index.
        throw out_of_range("Invalid column index in TTable::GetCell");
    }

    if (rowIndex >= GetNumRows())
    {
        // Wrong row index.
        throw out_of_range("Invalid row index in TTable::GetCell");
    }

    return((*_data[colIndex])[IntRowIndex(rowIndex)]);

}


int TTable::Write(FileNavigator* fileNav, int& size)
{

    int err;
    Word ret = 0, place = 0;


    // Write the number of rows
    // VLAD - THIS MAY NOT BE NECESSARY, SINCE IT CAN BE OBTAINED FROM
    // COLUMN LENGTH
    err = fileNav->WriteUWord(GetNumRows(), ret);
    if (err) fileNav->PrintError(err);

    // Write the number of columns
    err = fileNav->WriteUWord(_data.size(), place);
    if (err) fileNav->PrintError(err);

    // Write the cells
    vector<string> colToWrite;
    for (unsigned int colI = 0; colI < _data.size(); ++colI)
    {
        for (unsigned int rowI = 0; rowI < GetNumRows(); ++rowI)
        {
            colToWrite.push_back(operator()(rowI, colI));
        }
        err = fileNav->WriteStrings(colToWrite, place);
        colToWrite.clear();
    }

    size = place - ret + 1;

    return ret;

}


int TTable::Read(Word index, FileNavigator* fileNav)
{

    int err = 0;


    // unsigned int numRows = fileNav->GetUWord(index, err);
    // This is not used and is not read, only skipped
    index++;

    unsigned int numColumns = fileNav->GetUWord(index, err);
    index++;

    for (unsigned int colI = 0; colI < numColumns; ++colI)
    {
        vector<string> tmpCol;
        fileNav->GetStrings(tmpCol, index, err);
        index++;
        InsertColumn(colI, tmpCol);
    }

    return(index);

}


#ifndef IS_TABLE_ROW_DELETE_AS_REMOVE
void TTable::EnlargeRowMap(const unsigned int numRows)
{

    unsigned int firstIndex = _rowMap.size();

    for (unsigned int rowIndex = firstIndex;
      rowIndex < (firstIndex + numRows); rowIndex++)
    {
        _rowMap.push_back(rowIndex);
    }

}


void TTable::ReduceRowMap(const unsigned int numRows)
{

    _rowMap.erase(_rowMap.begin() + numRows, _rowMap.end());

}


void TTable::MarkRowDeleted(const unsigned int rowIndex)
{

    unsigned int saveRowIndex = _rowMap[rowIndex];

    unsigned int index;

    for (index = rowIndex; index < (_numRows - 1); index++)
    {
        _rowMap[index] = _rowMap[index + 1];
    }

    _rowMap[index] = saveRowIndex;

    ++_numDelRows;

}


void TTable::UnMarkRowDeleted(const unsigned int rowIndex)
{

    unsigned int availRowIndex = _rowMap[_numRows - 1];

    unsigned int index;

    for (index = _numRows - 1; index > rowIndex; --index)
    {
        _rowMap[index] = _rowMap[index - 1];
    }

    _rowMap[index] = availRowIndex;

    --_numDelRows;

}
#endif

